# SUR - Project
# May 2025
# Radek Zobaník (xzoban02), Michal Zobaník (xzoban01)
# Model for classiffication with audio files

import librosa 
import numpy as np 
from glob import glob
from sklearn.mixture import GaussianMixture
import os
import pickle
from result import Result

class Audio_GMM:
    def __init__(self, result_file_name:str = "audio_gmm", model_location:str = "voice_gmm.pkl"):
        self.model = {}
        self.mfcc = 30
        self.result_file_name = result_file_name
        self.model_location = model_location

    def _convert_to_mfcc(self, dir_name):
        features = []
        for f in glob(dir_name + '/*.wav'):
            print('Processing file:', f)
            y, sr = librosa.load(f, sr=None)
            mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=self.mfcc)
            delta = librosa.feature.delta(mfcc)
            delta2 = librosa.feature.delta(mfcc, order=2)
            parse = np.vstack([mfcc, delta, delta2])
            parse = parse.T
            features.append(parse)
        return np.vstack(features)

    def train(self, train_data_location: str, validate_data_location:str):
        for i in range(1,32):
            files = train_data_location + f'/{i}'
            print(f'GMM - speaker {i}')
            features = self._convert_to_mfcc(files)  
            gmm = GaussianMixture(n_components=24, covariance_type='diag', n_init=8, reg_covar=1e-2)
            gmm.fit(features)
            self.model[i] = gmm

        with open(self.model_location, "wb") as f:
            pickle.dump(self.model, f)

        self.validate(validate_data_location)

    def _load_model(self):
        with open(self.model_location, "rb") as f:
            self.model = pickle.load(f)

    def _predict(self, test_file: str):
        y, sr = librosa.load(test_file, sr=None)
        mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=self.mfcc)
        delta = librosa.feature.delta(mfcc)
        delta2 = librosa.feature.delta(mfcc, order=2)
        parse = np.vstack([mfcc, delta, delta2])
        parse = parse.T
        prob = {}

        for id, gmm in self.model.items():
            score = gmm.score(parse)
            prob[id] = score

        return prob
    
    def validate(self, test_data_location: str):
        correct = 0
        total = 0

        for i in range(1,32):
            files = test_data_location + f'/{i}'
            wav_files = glob(os.path.join(files, '*.wav'))
            
            for file in wav_files:
                prob = self._predict(file)
                pred_speaker_id = max(prob, key=prob.get)
                if pred_speaker_id == i:
                    correct += 1
                total += 1

        accuracy = correct / total
        print(f"Validation accuracy:  {accuracy * 100:.2f}%")

    def classification(self, data_location: str) -> list[Result]:
        if len(self.model) == 0:
            self._load_model()

        results = []
        wav_files = glob(os.path.join(data_location, '*.wav'))
            
        for file in wav_files:
            name = os.path.splitext(os.path.basename(file))[0]
            prob = self._predict(file)
            pred_speaker_id = max(prob, key=prob.get)
            prob_parse = np.full(31, 0)
            for i in range(1,32):
                prob_parse[i - 1] = prob[i]
            
            results.append(Result(name,pred_speaker_id,prob_parse))
        return results
