# SUR - Project
# May 2025
# Michal Zobaník (xzoban01), Radek Zobaník (xzoban02)
# Model for classiffication with audio files


import os
import keras
from keras import models
from keras import layers
from keras import optimizers
from keras import callbacks
from keras import utils
import numpy as np
from ray import tune

from result import Result

param_tune = False # should train model or tune model params
model_params = { # params of used model
            "conv1_filters": 16,
            "conv2_filters": 32,
            "extra_conv_blocks": 3,
            "conv3_filters_0": 64,
            "conv3_filters_1": 64,
            "conv3_filters_2": 64,
            "dense_1_units": 32,
            "dense_2_units": 64,
            "dropout_1": 0.5,
            "dropout_2": 0.5,
            "learning_rate": 0.001,
            "amsgrad": True,
            "dense_2_use": True,
            "dense_1_use": True,
            "conv_1_use": True,
            "conv_2_use": True
        }

class Image_cnn:
    def __init__(self, result_file_name:str = "image_cnn", model_location:str = "image_cnn_model"):
        self.result_file_name = result_file_name
        self.model_location = model_location
        self.model = None
        self.batch_size = 64
        self.epochs = 500
        self.images_per_epoch = 1111
        self.train_data_location = None
        self.valid_data_location = None

    def train(self, train_data_location:str, valid_data_location:str):
        if param_tune:
            # find best hyper params for model
            script_dir = os.path.dirname(os.path.abspath(__file__))
            parent_dir = os.path.dirname(script_dir)
            train_data_path = os.path.join(parent_dir, train_data_location)
            eval_data_path = os.path.join(parent_dir, valid_data_location)
            self._tune_model(train_data_path, eval_data_path)
        else:
            # prepare dataset and model
            (train_dataset, valid_dataset) = self._load_dataset(train_data_location, valid_data_location)
            model = self._get_model_tune(model_params)

            # callbacks for model training
            save_best = callbacks.ModelCheckpoint(self.model_location, save_best_only=True, monitor="val_accuracy")

            # train model
            history = model.fit(train_dataset, 
                            epochs = self.epochs, 
                            steps_per_epoch = self.images_per_epoch // self.batch_size,
                            validation_data = valid_dataset,
                            verbose = 2,
                            callbacks=[save_best])

    def classification(self, data_location:str) -> list[Result]:
        if self.model is None:
            self._load_model()

        # find images in folder
        images = []
        image_names = []
        for image_name in os.listdir(data_location):
            if image_name.endswith(".png"):
                image_path = os.path.join(data_location, image_name)
                image = utils.load_img(image_path)
                images.append(utils.img_to_array(image))
                image_names.append(image_name.replace(".png", ""))

        # use models
        model_result = self.model.predict(np.array(images))
        log_results = np.log(model_result + 1e-9)
        
        # parse results
        result = []
        for i in range(0, len(images)):
            result_i = log_results[i]
            file_name = image_names[i]
            chosen_class = np.argmax(result_i) + 1

            result.append(Result(file_name, chosen_class, result_i))

        return result

    def _load_model(self):
        self.model = keras.saving.load_model(self.model_location)

    def _get_data_augmentation_model(self):
        data_augment = models.Sequential(
            [
                layers.RandomFlip("horizontal"),
                layers.RandomRotation(0.05),
                layers.RandomZoom(0.1),
                layers.RandomTranslation(0.1, 0.1),
                layers.RandomBrightness(0.1),
                layers.RandomContrast(0.1),
                layers.RandomGaussianBlur(0.1),
                layers.RandomShear(0.1)
            ]
        )

        return data_augment

    def _load_dataset(self, train_data_location:str, valid_data_location:str):
        train_dataset = utils.image_dataset_from_directory(train_data_location,
                                                           batch_size=self.batch_size,
                                                           image_size=(80, 80)).repeat()
        
        data_augment = self._get_data_augmentation_model()
        train_dataset = train_dataset.map(lambda x, y: (data_augment(x), y))
                                                           
        validation_dataset = utils.image_dataset_from_directory(valid_data_location,
                                                           batch_size=self.batch_size,
                                                           image_size=(80, 80))

        return (train_dataset, validation_dataset)

    def _get_model_tune(self, config):
        model = models.Sequential()

        model.add(layers.Input((80,80,3)))
        model.add(layers.Rescaling(scale=1./255))

        if config["conv_1_use"]:
            model.add(layers.Conv2D(filters=config["conv1_filters"], kernel_size=(3,3), padding='Same', activation='relu'))
            model.add(layers.MaxPooling2D(pool_size=(3,3)))
            model.add(layers.BatchNormalization())

        if config["conv_2_use"]:
            model.add(layers.Conv2D(filters=config["conv2_filters"], kernel_size=(3,3), padding='Same', activation='relu'))
            model.add(layers.MaxPooling2D(pool_size=(2, 2)))
            model.add(layers.BatchNormalization())

        for i in range(config["extra_conv_blocks"]):
            model.add(layers.Conv2D(filters=config[f"conv3_filters_{i}"], kernel_size=(3,3), padding='same', activation='relu'))

        if config["extra_conv_blocks"] > 0:
            model.add(layers.MaxPooling2D(pool_size=(2,2)))
            model.add(layers.BatchNormalization())

        model.add(layers.Flatten())
        if config["dense_1_use"]:
            model.add(layers.Dense(config["dense_1_units"],activation='relu'))
            model.add(layers.Dropout(config["dropout_1"]))

        if config["dense_2_use"]:
            model.add(layers.Dense(config["dense_2_units"],activation='relu'))
            model.add(layers.Dropout(config["dropout_2"]))
        model.add(layers.Dense(31,activation='softmax'))

        optimizer = optimizers.Adam(learning_rate=config["learning_rate"], amsgrad=config["amsgrad"])
        model.compile(optimizer = optimizer, loss='sparse_categorical_crossentropy', metrics=["accuracy"])

        return model
    
    def _tune_train(self, config):
        model = self._get_model_tune(config)
        (train_dataset, valid_dataset) = self._load_dataset(self.train_data_location, self.valid_data_location)

        callbacks = []

        history = model.fit(train_dataset, 
                        epochs = self.epochs, 
                        steps_per_epoch = self.images_per_epoch // self.batch_size,
                        validation_data = valid_dataset,
                        verbose = 2,
                        callbacks=callbacks)

        val_acc = max(history.history["val_accuracy"])
        tune.report({"val_accuracy": val_acc})

    def _tune_model(self, train_data_location:str, valid_data_location:str):
        self.train_data_location = train_data_location
        self.valid_data_location = valid_data_location
        self.epochs = 111
        self.images_per_epoch = 222

        search_space = {
            "conv1_filters": tune.choice([16, 32, 64]),
            "conv2_filters": tune.choice([16, 32, 64]),
            "extra_conv_blocks": 3,
            "conv3_filters_0": tune.choice([16, 32, 64, 128]),
            "conv3_filters_1": tune.choice([16, 32, 64, 128]),
            "conv3_filters_2": tune.choice([16, 32, 64, 128]),
            "dense_1_units": tune.choice([16, 32, 64]),
            "dense_2_units": tune.choice([16, 32, 64]),
            "dropout_1": 0.5,
            "dropout_2": 0.5,
            "learning_rate": tune.choice([0.1, 0.01, 0.001]),
            "amsgrad": True,
            "dense_2_use": True,
            "dense_1_use": True,
            "conv_1_use": True,
            "conv_2_use": True
        }

        tuner = tune.Tuner(
        self._tune_train,
        param_space=search_space,
        tune_config=tune.TuneConfig(
            metric="val_accuracy",
            mode="max",
            num_samples=100, 
            max_concurrent_trials=1
            ),
        )

        results = tuner.fit()

        print("\n\nRESULTS:")
        for i, result in enumerate(results):
            config = result.config
            val_acc = result.metrics.get("val_accuracy")
            print(f"Trial {i}: Accuracy = {val_acc:.4f}, Config = {config}\n")

        print("\nBEST:\n")

        best_result = results.get_best_result("val_accuracy", "max")
        print("Best config:", best_result.config)
